"""
Helper, Property and Static method tests for pyunigps.UNIMessage

Created on 26 Jan 2026

*** NB: must be saved in UTF-8 format ***

@author: semuadmin
"""

# pylint: disable=line-too-long, invalid-name, missing-docstring, no-member

import os
import unittest
from datetime import datetime, timezone
import pyunigps.unitypes_core as unt
import pyunigps.exceptions as une
from pyunigps.unitypes_core import CV, UNI_MSGIDS
from pyunigps.unihelpers import (
    calc_crc,
    escapeall,
    att2idx,
    att2name,
    attsiz,
    get_bits,
    val2bytes,
    bytes2val,
    nomval,
    key_from_val,
    timeinfo2bytes,
    timeinfo2vals,
    utc2wnotow,
)


class StaticTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        dirname = os.path.dirname(__file__)

    def tearDown(self):
        pass

    def testcrc(self):

        MSG1 = b"\xaa\x44\x5b\x00\x12\x00\x05\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\x00\x11\x22\x33\x44\x55\x66\x11\x22\x33\x44\x55"
        res = calc_crc(MSG1)
        # print(escapeall(res))
        self.assertEqual(res, b"\xc6\xe9\x40\x60")

        MSG2 = b"\xaa\x44\x5b\x00\x14\x00\x07\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\x00\x11\x22\x33\x44\x55\x66\x11\x22\x33\x44\x55\x66\x77"
        res = calc_crc(MSG2)
        # print(escapeall(res))
        self.assertEqual(res, b"\x70\x19\x8f\x95")

    def testVal2Bytes(self):  # test conversion of value to bytes
        INPUTS = [
            (2345, unt.U2),
            (b"\x44\x55", unt.X2),
            (23.12345678, unt.R4),
            (-23.12345678912345, unt.R8),
            ("test1234", unt.C8),
        ]
        EXPECTED_RESULTS = [
            b"\x29\x09",
            b"\x44\x55",
            b"\xd7\xfc\xb8\x41",
            b"\x1f\xc1\x37\xdd\x9a\x1f\x37\xc0",
            b"test1234",
        ]
        for i, inp in enumerate(INPUTS):
            val, att = inp
            res = val2bytes(val, att)
            self.assertEqual(res, EXPECTED_RESULTS[i])

    def testVal2BytesInvalid(self):
        with self.assertRaisesRegex(une.UNITypeError, "Unknown attribute type Y002"):
            res = val2bytes(1234, "Y002")

    def testBytes2Val(self):  # test conversion of bytes to value
        INPUTS = [
            (b"\x29\x09", unt.U2),
            (b"\x44\x55", unt.X2),
            (b"\xd7\xfc\xb8\x41", unt.R4),
            (b"\x1f\xc1\x37\xdd\x9a\x1f\x37\xc0", unt.R8),
            (b"test1234", unt.C8),
        ]
        EXPECTED_RESULTS = [
            2345,
            b"\x44\x55",
            23.12345678,
            -23.12345678912345,
            "test1234",
        ]
        for i, inp in enumerate(INPUTS):
            valb, att = inp
            res = bytes2val(valb, att)
            if att == unt.R4:
                self.assertAlmostEqual(res, EXPECTED_RESULTS[i], 6)
            elif att == unt.R8:
                self.assertAlmostEqual(res, EXPECTED_RESULTS[i], 14)
            else:
                self.assertEqual(res, EXPECTED_RESULTS[i])

    def testBytes2ValInvalid(self):
        with self.assertRaisesRegex(une.UNITypeError, "Unknown attribute type Y002"):
            res = bytes2val(b"\x12\x34", "Y002")

    def testNomval(self):  # test conversion of value to bytes
        INPUTS = [
            unt.U2,
            unt.X2,
            unt.R4,
            unt.R8,
            unt.C8,
        ]
        EXPECTED_RESULTS = [
            0,
            b"\x00\x00",
            0.0,
            0.0,
            "        ",
        ]
        for i, att in enumerate(INPUTS):
            res = nomval(att)
            self.assertEqual(res, EXPECTED_RESULTS[i])

    def testNomValInvalid(self):
        with self.assertRaisesRegex(une.UNITypeError, "Unknown attribute type Y002"):
            res = nomval("Y002")

    def testgetbits(self):
        INPUTS = [
            (b"\x89", 192),
            (b"\xc9", 3),
            (b"\x89", 9),
            (b"\xc9", 9),
            (b"\x18\x18", 8),
            (b"\x18\x20", 8),
        ]
        EXPECTED_RESULTS = [2, 1, 9, 9, 1, 0]
        for i, (vb, mask) in enumerate(INPUTS):
            vi = get_bits(vb, mask)
            self.assertEqual(vi, EXPECTED_RESULTS[i])

    def testattsiz(self):  # test attsiz
        self.assertEqual(attsiz(CV), -1)
        self.assertEqual(attsiz("C032"), 32)

    def testatt2idx(self):  # test att2idx
        EXPECTED_RESULT = [4, 16, 101, 0, (3, 6), 0]
        atts = ["svid_04", "gnssId_16", "cno_101", "gmsLon", "gnod_03_06", "dodgy_xx"]
        for i, att in enumerate(atts):
            res = att2idx(att)
            # print(res)
            self.assertEqual(res, EXPECTED_RESULT[i])

    def testatt2name(self):  # test att2name
        EXPECTED_RESULT = ["svid", "gnssId", "cno", "gmsLon"]
        atts = ["svid_04", "gnssId_16", "cno_101", "gmsLon"]
        for i, att in enumerate(atts):
            res = att2name(att)
            # print(res)
            self.assertEqual(res, EXPECTED_RESULT[i])

    def testescapeall(self):
        EXPECTED_RESULT = "b'\\x68\\x65\\x72\\x65\\x61\\x72\\x65\\x73\\x6f\\x6d\\x65\\x63\\x68\\x61\\x72\\x73'"
        val = b"herearesomechars"
        res = escapeall(val)
        print(res)
        self.assertEqual(res, EXPECTED_RESULT)

    def testkeyfromval(self):
        res = key_from_val(UNI_MSGIDS, "GLOEPH")
        self.assertEqual(res, 107)

    def testkeyfromvalinvalid(self):
        with self.assertRaisesRegex(KeyError, "No key found for value XXXX"):
            res = key_from_val(UNI_MSGIDS, "XXXX")

    def testutc2wnotow(self):
        dat = datetime(2026, 1, 28, 9, 34, 12, 234000, tzinfo=timezone.utc)
        wno, tow = utc2wnotow(dat)
        print(wno, tow)
        self.assertEqual((wno, tow), (2403, 293652234))

    def testtimeinfo2bytes(self):
        t = timeinfo2bytes(wno=2406, tow=34675834)
        # print(t)
        self.assertEqual(t, b'\x01\x00f\tz\x1c\x11\x02\x00\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(len(t), 16)
        t = timeinfo2bytes(1,1, 2402, 34675834,1,18,23)
        # print(t)
        self.assertEqual(t, b'\x01\x01b\tz\x1c\x11\x02\x01\x00\x00\x00\x00\x12\x17\x00')
        self.assertEqual(len(t), 16)
    
    def testtimeinfo2vals(self):
        v = timeinfo2vals(b'\x01\x00f\tz\x1c\x11\x02\x00\x00\x00\x00\x00\x00\x00\x00')
        # print(v)
        self.assertEqual(v,(1, 0, 2406, 34675834, 0, 0, 0))
        v = timeinfo2vals(b'\x01\x01b\tz\x1c\x11\x02\x01\x00\x00\x00\x00\x12\x17\x00')
        # print(v)
        self.assertEqual(v,(1,1, 2402, 34675834,1,18,23))

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
UNImessage.py

Main UNI Message Protocol Class.

Created on 26 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2020
:license: BSD 3-Clause
"""

# pylint: disable=too-many-positional-arguments, too-many-locals, too-many-arguments

import struct
from types import NoneType

from pyunigps.exceptions import UNIMessageError, UNITypeError
from pyunigps.unihelpers import (
    attsiz,
    bytes2val,
    calc_crc,
    escapeall,
    nomval,
    timeinfo2bytes,
    utc2wnotow,
    val2bytes,
)
from pyunigps.unitypes_core import (
    GET,
    POLL,
    SCALROUND,
    SET,
    U1,
    U2,
    UNI_HDR,
    UNI_MSGIDS,
)
from pyunigps.unitypes_get import UNI_PAYLOADS_GET
from pyunigps.unitypes_poll import UNI_PAYLOADS_POLL
from pyunigps.unitypes_set import UNI_PAYLOADS_SET


class UNIMessage:
    """UNI Message Class."""

    def __init__(
        self,
        msgid: int,
        length: int | NoneType = None,
        cpuidle: int = 0,
        timeref: int = 0,
        timestatus: int = 0,
        wno: int | NoneType = None,
        tow: int | NoneType = None,
        version: int = 0,
        leapsecond: int = 0,
        delay: int = 0,
        checksum: bytes | NoneType = None,
        msgmode: int = GET,
        parsebitfield: bool = True,
        **kwargs,
    ):
        """
        If no keyword parms are passed, the payload is taken to be empty.

        If 'payload' is passed as a keyword parm, this is taken to contain the complete
        payload as a sequence of bytes; any other keyword parms are ignored.

        Otherwise, any named attributes will be assigned the value given, all others will
        be assigned a nominal value according to type.

        :param msgid: msgid
        :param int | NoneType length: length (will be derived if None)
        :param int cpuidle: header cpuidle
        :param int timeref: header timeref
        :param int timestatus: header timestatus
        :param int | NoneType wno: header week number
        :param int | NoneType tow: header time of week
        :param int version: header version
        :param int leapsecond: header leapsecond
        :param int delay: header delay
        :param bytes | NoneType checksum: CRC (will be derived if None)
        :param int msgmode: message mode (0 = GET, 1 = SET, 2 = POLL)
        :param bool parsebitfield: 0 = parse as bytes, 1 = parse as individual bits
        :param kwargs: optional keywords representing payload attributes
        :raises: UNITypeError, UNIMessageError
        """

        # object is mutable during initialisation only
        super().__setattr__("_immutable", False)
        self.cpuidle = cpuidle
        self._length = length
        self._checksum = checksum  # bytes
        self._msgid = msgid
        self.cpuidle = cpuidle
        self.timeref = timeref
        self.timestatus = timestatus
        if wno is None or tow is None:  # default to now
            wno, tow = utc2wnotow()
        self.wno = wno
        self.tow = tow
        self.version = version
        self.leapsecond = leapsecond
        self.delay = delay
        # serialized version of header time info
        self._timeinfob = timeinfo2bytes(
            self.timeref,
            self.timestatus,
            self.wno,
            self.tow,
            self.version,
            self.leapsecond,
            self.delay,
        )
        self._mode = msgmode
        self._payload = b""
        self._parsebf = parsebitfield  # parsing bitfields Y/N?

        if msgmode not in (GET, SET, POLL):
            raise UNIMessageError(f"Invalid msgmode {msgmode} - must be 0, 1 or 2")

        self._do_attributes(**kwargs)

        self._immutable = True  # once initialised, object is immutable

    def _do_attributes(self, **kwargs):
        """
        Populate UNIMessage from named attribute keywords.
        Where a named attribute is absent, set to a nominal value (zeros or blanks).

        :param kwargs: optional payload key/value pairs
        :raises: UNITypeError

        """

        offset = 0  # payload offset in bytes
        index = []  # array of (nested) group indices

        try:
            if len(kwargs) == 0:  # if no kwargs, assume null payload
                self._payload = None
            else:
                self._payload = kwargs.get("payload", b"")
                pdict = self._get_dict(**kwargs)  # get appropriate payload dict
                for anam in pdict:  # process each attribute in dict
                    offset, index = self._set_attribute(
                        anam, pdict, offset, index, **kwargs
                    )
            self._do_len_checksum()

        except (
            AttributeError,
            struct.error,
            TypeError,
            ValueError,
        ) as err:
            raise UNITypeError(
                (
                    f"Incorrect type for attribute '{anam}' "
                    f"in {['GET', 'SET', 'POLL'][self._mode]} message class {self.identity}"
                )
            ) from err
        except (OverflowError,) as err:
            raise UNITypeError(
                (
                    f"Overflow error for attribute '{anam}' "
                    f"in {['GET', 'SET', 'POLL'][self._mode]} message class {self.identity}"
                )
            ) from err

    def _set_attribute(
        self, anam: str, pdict: dict, offset: int, index: list, **kwargs
    ) -> tuple:
        """
        Recursive routine to set individual or grouped payload attributes.

        :param str anam: attribute name
        :param dict pdict: dict representing payload definition
        :param int offset: payload offset in bytes
        :param list index: repeating group index array
        :param kwargs: optional payload key/value pairs
        :return: (offset, index[])
        :rtype: tuple

        """

        adef = pdict[anam]  # get attribute definition
        if isinstance(
            adef, tuple
        ):  # repeating group of attributes or subdefined bitfield
            numr, _ = adef
            if numr[0] == "X":  # bitfield
                if self._parsebf:  # if we're parsing bitfields
                    offset, index = self._set_attribute_bitfield(
                        adef, offset, index, **kwargs
                    )
                else:  # treat bitfield as a single byte array
                    offset = self._set_attribute_single(
                        anam, numr, offset, index, **kwargs
                    )
            else:  # repeating group of attributes
                offset, index = self._set_attribute_group(adef, offset, index, **kwargs)
        else:  # single attribute
            offset = self._set_attribute_single(anam, adef, offset, index, **kwargs)

        return (offset, index)

    def _set_attribute_group(
        self, adef: tuple, offset: int, index: list, **kwargs
    ) -> tuple:
        """
        Process (nested) group of attributes.

        :param tuple adef: attribute definition - tuple of (num repeats, attribute dict)
        :param int offset: payload offset in bytes
        :param list index: repeating group index array
        :param kwargs: optional payload key/value pairs
        :return: (offset, index[])
        :rtype: tuple

        """

        index.append(0)  # add a (nested) group index
        anam, gdict = adef  # attribute signifying group size, group dictionary
        # derive or retrieve number of items in group
        if isinstance(anam, int):  # fixed number of repeats
            gsiz = anam
        elif anam == "None":  # number of repeats 'variable by size'
            gsiz = self._calc_num_repeats(gdict, self._payload, offset, 0)
        else:  # number of repeats is defined in named attribute
            gsiz = getattr(self, anam)
        # recursively process each group attribute,
        # incrementing the payload offset and index as we go
        for i in range(gsiz):
            index[-1] = i + 1
            for key1 in gdict:
                offset, index = self._set_attribute(
                    key1, gdict, offset, index, **kwargs
                )

        index.pop()  # remove this (nested) group index

        return (offset, index)

    def _set_attribute_single(
        self, anam: str, adef: object, offset: int, index: list, **kwargs
    ) -> int:
        """
        Set individual attribute value, applying scaling where appropriate.

        :param str anam: attribute keyword
        EITHER
        :param str adef: attribute definition string e.g. 'U002'
        OR
        :param list adef: if scaled, list of [attribute type string, scaling factor float]
        :param int offset: payload offset in bytes
        :param list index: repeating group index array
        :param kwargs: optional payload key/value pairs
        :return: offset
        :rtype: int

        """
        # pylint: disable=no-member

        # if attribute is scaled
        ares = 1
        if isinstance(adef, list):
            ares = adef[1]  # attribute resolution (i.e. scaling factor)
            adef = adef[0]  # attribute definition

        # if attribute is part of a (nested) repeating group, suffix name with index
        anami = anam
        for i in index:  # one index for each nested level
            if i > 0:
                anami += f"_{i:02d}"

        # determine attribute size (bytes) - some attributes have
        # variable length, depending on
        # - multiple of value of preceding attribute
        # - payload length - offset
        asiz = attsiz(adef)

        # if payload keyword has been provided,
        # use the appropriate offset of the payload
        if "payload" in kwargs:
            valb = self._payload[offset : offset + asiz]
            if ares == 1:
                val = bytes2val(valb, adef)
            else:
                val = round(bytes2val(valb, adef) * ares, SCALROUND)
        else:
            # if individual keyword has been provided,
            # set to provided value, else set to
            # nominal value
            val = kwargs.get(anami, nomval(adef))
            if ares == 1:
                valb = val2bytes(val, adef)
            else:
                valb = val2bytes(int(val / ares), adef)
            self._payload += valb

        setattr(self, anami, val)

        return offset + asiz

    def _set_attribute_bitfield(
        self, atyp: str, offset: int, index: list, **kwargs
    ) -> tuple:
        """
        Parse bitfield attribute (type 'X').

        :param str atyp: attribute type e.g. 'X002'
        :param int offset: payload offset in bytes
        :param list index: repeating group index array
        :param kwargs: optional payload key/value pairs
        :return: (offset, index[])
        :rtype: tuple

        """
        # pylint: disable=no-member

        btyp, bdict = atyp  # type of bitfield, bitfield dictionary
        bsiz = attsiz(btyp)  # size of bitfield in bytes
        bfoffset = 0

        # if payload keyword has been provided,
        # use the appropriate offset of the payload
        if "payload" in kwargs:
            bitfield = int.from_bytes(self._payload[offset : offset + bsiz], "little")
        else:
            bitfield = 0

        # process each flag in bitfield
        for key, keyt in bdict.items():
            bitfield, bfoffset = self._set_attribute_bits(
                bitfield, bfoffset, key, keyt, index, **kwargs
            )

        # update payload
        if "payload" not in kwargs:
            self._payload += bitfield.to_bytes(bsiz, "little")

        return (offset + bsiz, index)

    def _set_attribute_bits(
        self,
        bitfield: int,
        bfoffset: int,
        key: str,
        keyt: str,
        index: list,
        **kwargs,
    ) -> tuple:
        """
        Set individual bit flag from bitfield.

        :param int bitfield: bitfield
        :param int bfoffset: bitfield offset in bits
        :param str key: attribute key name
        :param str keyt: key type e.g. 'U001'
        :param list index: repeating group index array
        :param kwargs: optional payload key/value pairs
        :return: (bitfield, bfoffset)
        :rtype: tuple

        """
        # pylint: disable=no-member

        # if attribute is part of a (nested) repeating group, suffix name with index
        keyr = key
        for i in index:  # one index for each nested level
            if i > 0:
                keyr += f"_{i:02d}"

        atts = attsiz(keyt)  # determine flag size in bits

        if "payload" in kwargs:
            val = (bitfield >> bfoffset) & ((1 << atts) - 1)
        else:
            val = kwargs.get(keyr, 0)
            bitfield = bitfield | (val << bfoffset)

        if key[0:8] != "reserved":  # don't bother to set reserved bits
            setattr(self, keyr, val)
        return (bitfield, bfoffset + atts)

    def _do_len_checksum(self):
        """
        Calculate and format payload length and checksum as bytes,
        if not passed as input arguments.
        """

        payload = b"" if self._payload is None else self._payload
        if self._length is None:
            self._length = len(payload)
        if self._checksum is None:
            cpuidleb = val2bytes(self.cpuidle, U1)
            msgidb = val2bytes(self._msgid, U2)
            lenb = val2bytes(self._length, U2)
            self._checksum = calc_crc(
                UNI_HDR + cpuidleb + msgidb + lenb + self._timeinfob + payload
            )

    def _get_dict(self, **kwargs) -> dict:  # pylint: disable=unused-argument
        """
        Get payload dictionary corresponding to message mode (GET/SET/POLL)

        :param kwargs: optional payload key/value pairs
        :return: dictionary representing payload definition
        :rtype: dict

        """

        try:
            if self._mode == POLL:
                pdict = UNI_PAYLOADS_POLL[self.identity]
            elif self._mode == SET:
                pdict = UNI_PAYLOADS_SET[self.identity]
            else:
                # Unknown GET message, parsed to nominal definition
                if self.identity[-7:] == "NOMINAL":
                    pdict = {}
                else:
                    pdict = UNI_PAYLOADS_GET[self.identity]
            return pdict
        except KeyError as err:
            mode = ["GET", "SET", "POLL"][self._mode]
            raise UNIMessageError(
                f"Unknown message type {self._msgid}, mode {mode}"
            ) from err

    def _calc_num_repeats(
        self, attd: dict, payload: bytes, offset: int, offsetend: int = 0
    ) -> int:
        """
        Deduce number of items in 'variable by size' repeating group by
        dividing length of remaining payload by length of group.

        This is predicated on there being only one such repeating group
        per message payload, which is true for all currently supported types.

        :param dict attd: grouped attribute dictionary
        :param bytes payload : raw payload
        :param int offset: number of bytes in payload before repeating group
        :param int offsetend: number of bytes in payload after repeating group
        :return: number of repeats
        :rtype: int

        """

        lenpayload = len(payload) - offset - offsetend
        lengroup = 0
        for _, val in attd.items():
            if isinstance(val, tuple):
                val, _ = val
            lengroup += attsiz(val)
        return int(lenpayload / lengroup)

    def __str__(self) -> str:
        """
        Human readable representation.

        :return: human readable representation
        :rtype: str

        """

        umsg_name = self.identity
        if self.payload is None:
            return f"<UNI({umsg_name})>"
        if self.identity[-7:] == "NOMINAL":
            return f"<UNI({umsg_name}, payload={escapeall(self._payload)})>"

        stg = f"<UNI({umsg_name}, "
        for i, att in enumerate(self.__dict__):
            if att[0] != "_":  # only show public attributes
                val = self.__dict__[att]
                # escape all byte chars unless they're
                # intended to be character strings
                if isinstance(val, bytes):
                    val = escapeall(val)
                stg += att + "=" + str(val).strip(" ")
                if i < len(self.__dict__) - 1:
                    stg += ", "
        stg += ")>"

        return stg

    def __repr__(self) -> str:
        """
        Machine readable representation.

        eval(repr(obj)) = obj

        :return: machine readable representation
        :rtype: str

        """

        rep = (
            f"UNIMessage({self._msgid}, {self._length}, {self.cpuidle}, {self.timeref}, "
            f"{self.timestatus},{self.wno}, {self.tow}, {self.version}, {self.leapsecond}, "
            f"{self.delay}, {self.checksum}, {self._mode}, {self._parsebf}"
        )
        if self._payload is not None:
            rep += f", payload={self._payload})"
        return rep

    def __setattr__(self, name, value):
        """
        Override setattr to make object immutable after instantiation.

        :param str name: attribute name
        :param object value: attribute value
        :raises: UNIMessageError

        """

        if self._immutable:
            raise UNIMessageError(
                f"Object is immutable. Updates to {name} not permitted after initialisation."
            )

        super().__setattr__(name, value)

    def serialize(self) -> bytes:
        """
        Serialize message.

        :return: serialized output
        :rtype: bytes

        """

        cpuidleb = val2bytes(self.cpuidle, U1)
        msgidb = val2bytes(self._msgid, U2)
        lenb = val2bytes(self._length, U2)
        ser = UNI_HDR + cpuidleb + msgidb + lenb + self._timeinfob
        payloadb = b"" if self._payload is None else self._payload
        return ser + payloadb + self._checksum

    @property
    def identity(self) -> str:
        """
        Returns message identity in plain text form.

        If the message is unrecognised, the message is parsed
        to a nominal payload definition UNI-NOMINAL and
        the term 'NOMINAL' is appended to the identity.

        :return: message identity e.g. 'OBSVMCMP'
        :rtype: str

        """

        try:
            umsg_name = UNI_MSGIDS[self._msgid]
        except KeyError:
            # unrecognised Unicore message, parsed to UNI-NOMINAL definition
            umsg_name = f"{int.from_bytes(self._msgid, 'little'):02x}-NOMINAL"
        return umsg_name

    @property
    def checksum(self) -> bytes:
        """
        CRC checksum getter.

        :return: CRC as bytes
        :rtype: bytes

        """

        return self._checksum

    @property
    def payload(self) -> bytes:
        """
        Payload getter - returns the raw payload bytes.

        :return: raw payload as bytes
        :rtype: bytes

        """

        return self._payload

    @property
    def msgmode(self) -> int:
        """
        Message mode getter.

        :return: msgmode as integer
        :rtype: int

        """

        return self._mode



"""
UNI Protocol core globals, constants, datatypes and message identifiers.

Created on 26 Jan 2026

Information sourced from public domain Unicore UM980 Interface Specifications © 2023, Unicore
https://www.ardusimple.com/wp-content/uploads/2023/04/Unicore-Reference-Commands-Manual-For-N4-High-Precision-Products_V2_EN_R1.4-1.pdf

:author: semuadmin (Steve Smith)
"""

UNI_HDR = b"\xaa\x44\xb5"
"""UNI message header"""
GET = 0
"""GET (receive, response) message types"""
SET = 1
"""SET (command) message types"""
POLL = 2
"""POLL (query) message types"""
SETPOLL = 3
"""SETPOLL (SET or POLL) message types"""
VALNONE = 0
"""Do not validate checksum"""
VALCKSUM = 1
"""Validate checksum"""
NMEA_PROTOCOL = 1
"""NMEA Protocol"""
UNI_PROTOCOL = 2
"""UNI Protocol"""
RTCM3_PROTOCOL = 4
"""RTCM3 Protocol"""
ERR_RAISE = 2
"""Raise error and quit"""
ERR_LOG = 1
"""Log errors"""
ERR_IGNORE = 0
"""Ignore errors"""
SCALROUND = 12  # number of dp to round scaled attributes to

# **************************************************
# THESE ARE THE UNI PROTOCOL PAYLOAD ATTRIBUTE TYPES
# **************************************************
C8 = "C008"  # 8 byte haracter string
C10 = "C010"  # 10 byte character string
CV = "CXXX"  # variable length character string
R4 = "R004"  # single precision float 4 [-1*2^127,2^127]
R8 = "R008"  # double precision float 8 [-1*2^1023,2^1023]
S1 = "S001"  # signed char 1 [-128,127]
S2 = "S002"  # signed short int 2 [-32768,32767]
S4 = "S004"  # signed int 4 [-2147483648,2147483647]
S8 = "S008"  # signed long long int 8 [-2^63,2^63-1]
U1 = "U001"  # unsigned char 1 [0,255]
U2 = "U002"  # unsigned short int 2 [0,65535]
U3 = "U003"  # unsigned short int 3
U4 = "U004"  # unsigned int 4 [0,4294967295]
U5 = "U005"  # unsigned int 5
U6 = "U006"  # unsigned int 6
U8 = "U008"  # unsigned long long int 8 [0,2^64-1]
U10 = "U010"  # unsigned long long int 10
U15 = "U015"  # unsigned long long int 15
U16 = "U016"  # unsigned long long int 16
U17 = "U017"  # unsigned long long int 17
X1 = "X001"  # 8 bits field 1 Bit 7-0
X2 = "X002"  # 16 bits field 2 Bit 15-0
X4 = "X004"  # 32 bits field 4 Bit 31-0
X8 = "X008"  # 64 bits field 8 Bit 63-0
X61 = "X061"  # 61 bytes
X250 = "X250"  # 250 bytes

ATTTYPE = {
    "S": type(-1),
    "R": type(1.1),
    "U": type(1),
    "X": type(b"0"),
    "C": type("X"),
}
"""Permissible attribute types"""

# ***************************************************************************
# THESE ARE THE UNI PROTOCOL CORE MESSAGE IDENTITIES
# Payloads for each of these identities are defined in the unitypes_* modules
#
# Uses nominal msgids in the 65000 range for messages which are only
# available in ASCII format
# ***************************************************************************
UNI_MSGIDS = {
    17: "VERSION",  # Version and Authorization
    12: "OBSVM",  # Observation of the Main Antenna
    13: "OBSVH",  # Observation of the 2nd Antenna
    138: "OBSVMCMP",  # Compressed Observation of the Main Antenna
    139: "OBSVHCMP",  # Compressed Observation of the 2nd Antenna
    284: "OBSVBASE",  # Observation of the Base Station
    176: "BASEINFO",  # Base Station Information
    8: "GPSION",  # GPS Ionosphere Parameters
    21: "BD3ION",  # BDS-3 Ionosphere Parameters
    4: "BDSION",  # BDS Ionosphere Parameters
    9: "GALION",  # Galileo Ionosphere Parameters
    19: "GPSUTC",  # Conversion between GPS Time and UTC
    22: "BD3UTC",  # Conversion between BDS-3 Time and UTC
    2012: "BDSUTC",  # Conversion between BDS Time and UTC
    20: "GALUTC",  # Conversion between Galileo Time and UTC
    106: "GPSEPH",  # GPS Ephemeris
    110: "QZSSEPH",  # QZSS Ephemeris
    2999: "BD3EPH",  # BDS-3 Ephemeris
    108: "BDSEPH",  # BDS Ephemeris
    107: "GLOEPH",  # GLONASS Ephemeris
    109: "GALEPH",  # Galileo Ephemeris
    112: "IRNSSEPH",  # IRNSS Ephemeris
    11276: "AGRIC",  # Position, velocity, serial no, heading and baseline information
    1021: "PVTSLN",  # Position and Heading Information
    65001: "UNILOGLIST",  # Output Log List (no binary version)
    2118: "BESTNAV",  # Best Position and Velocity
    240: "BESTNAVXYZ",  # Best Position and Velocity in ECEF
    2119: "BESTNAVH",  # Best Position and Velocity (2nd Antenna)
    242: "BESTNAVXYZH",  # Best Position and Velocity in ECEF (2nd Antenna)
    1041: "BESTSAT",  # Satellites Used in Position Solution
    142: "ADRNAV",  # RTK Position and Velocity
    2117: "ADRNAVH",  # RTK Position and Velocity (2nd Antenna)
    1026: "PPPNAV",  # Position and Velocity of PPP
    46: "SPPNAV",  # Pseudorange Position and Velocity
    2116: "SPPNAVH",  # Pseudorange Position and Velocity (2nd Antenna)
    954: "STADOP",  # DOP of BESTNAV
    2122: "STADOPH",  # DOP of BESTNAVH (2nd Antenna)
    953: "ADRDOP",  # DOP of ADRNAV
    2121: "ADRDOPH",  # DOP of ADRNAVH (2nd Antenna)
    1025: "PPPDOP",  # DOP of PPPNAV
    173: "SPPDOP",  # DOP of SPPNAV
    2120: "SPPDOPH",  # DOP of SPPNAVH (2nd Antenna)
    2124: "SATSINFO",  # Satellite Information
    49: "BASEPOS",  # Position of the Base Station
    1042: "SATELLITE",  # Visible Satellites
    2115: "SATECEF",  # Satellite Coordinates in ECEF
    102: "RECTIME",  # Time Information
    972: "UNIHEADING",  # Heading Information
    1331: "UNIHEADING2",  # Multi-Rover Heading Information
    521: "HEADINGSTATUS",  # Heading Status
    509: "RTKSTATUS",  # RTK Solution Status
    512: "AGNSSSTATUS",  # AGNSS Status
    510: "RTCSTATUS",  # RTC Initialization Status
    511: "JAMSTATUS",  # Jamming Detection
    519: "FREQJAMSTATUS",  # Frequency Jamming Status
    2125: "RTCMSTATUS",  # RTCM Data Status
    218: "HWSTATUS",  # Hardware Status
    220: "AGC",  # Automatic Gain Control
    65002: "KSXT",  # Positioning and Heading Data Output (no binary format)
    1019: "INFOPART1",  # Read user-defined information in PART1
    1020: "INFOPART2",  # Read user-defined information in PART2
    520: "MSPOS",  # Best Position of Dual Antennas
    2318: "TROPINFO",  # Zenith Tropospheric Delay
    2302: "PPPB2BINFO1",  # Information Type 1
    2304: "PPPB2BINFO2",  # Information Type 2
    2306: "PPPB2BINFO3",  # Information Type 3
    2308: "PPPB2BINFO4",  # Information Type 4
    2310: "PPPB2BINFO5",  # Information Type 5
    2312: "PPPB2BINFO6",  # Information Type 6
    2314: "PPPB2BINFO7",  # Information Type 7
    2319: "E6MASKBLOCK",  # Mask Block
    2320: "E6ORBITBLOCK",  # Orbit Corrections Block
    2321: "E6CLOCKFULLBLOCK",  # Clock Full-Set Corrections Block
    2322: "E6CLOCKSUBBLOCK",  # Clock Subset Corrections Block
    2323: "E6CBIASBLOCK",  # Code Biases Block
    2324: "E6PBIASBLOCK",  # Phase Biases Block
    1316: "BSLNENUHD2",  # Heading2 Baseline in ENU Coordinate System
    1317: "BSLNXYZHD2",  # Heading2 Baseline in XYZ Coordinate System
    1333: "DOPHD2",  # DOP of Heading2
    # TODO remove after Alpha..
    65512: "TEST12",
    65514: "TEST14",
}



"""
Created on 6 Oct 2025

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2020
:license: BSD 3-Clause
"""

from pynmeagps import SocketWrapper

from pyunigps._version import __version__
from pyunigps.exceptions import (
    GNSSStreamError,
    ParameterError,
    UNIMessageError,
    UNIParseError,
    UNIStreamError,
    UNITypeError,
)
from pyunigps.unihelpers import *
from pyunigps.unimessage import UNIMessage
from pyunigps.unireader import UNIReader
from pyunigps.unitypes_core import *
from pyunigps.unitypes_get import *

version = __version__  # pylint: disable=invalid-name



"""
UNI Protocol query definitions.

Created on 26 Jan 2026

Information sourced from public domain Unicore UM980 Interface Specifications © 2023, Unicore
https://www.ardusimple.com/wp-content/uploads/2023/04/Unicore-Reference-Commands-Manual-For-N4-High-Precision-Products_V2_EN_R1.4-1.pdf

:author: semuadmin (Steve Smith)
"""

UNI_PAYLOADS_POLL = {}



"""
Collection of UNI helper methods which can be used
outside the UNIMessage or UNIReader classes.

Created on 6 Oct 2025

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2020
:license: BSD 3-Clause
"""

import struct
from datetime import datetime, timezone
from types import NoneType
from typing import Any

import pyunigps.exceptions as qge
from pyunigps.unitypes_core import ATTTYPE, U1, U2, U4

GPSEPOCH0 = datetime(1980, 1, 6, tzinfo=timezone.utc)
# ARC table for CRC calculation in calc_crc
CRCTABLE = [
    0x00000000,
    0x77073096,
    0xEE0E612C,
    0x990951BA,
    0x076DC419,
    0x706AF48F,
    0xE963A535,
    0x9E6495A3,
    0x0EDB8832,
    0x79DCB8A4,
    0xE0D5E91E,
    0x97D2D988,
    0x09B64C2B,
    0x7EB17CBD,
    0xE7B82D07,
    0x90BF1D91,
    0x1DB71064,
    0x6AB020F2,
    0xF3B97148,
    0x84BE41DE,
    0x1ADAD47D,
    0x6DDDE4EB,
    0xF4D4B551,
    0x83D385C7,
    0x136C9856,
    0x646BA8C0,
    0xFD62F97A,
    0x8A65C9EC,
    0x14015C4F,
    0x63066CD9,
    0xFA0F3D63,
    0x8D080DF5,
    0x3B6E20C8,
    0x4C69105E,
    0xD56041E4,
    0xA2677172,
    0x3C03E4D1,
    0x4B04D447,
    0xD20D85FD,
    0xA50AB56B,
    0x35B5A8FA,
    0x42B2986C,
    0xDBBBC9D6,
    0xACBCF940,
    0x32D86CE3,
    0x45DF5C75,
    0xDCD60DCF,
    0xABD13D59,
    0x26D930AC,
    0x51DE003A,
    0xC8D75180,
    0xBFD06116,
    0x21B4F4B5,
    0x56B3C423,
    0xCFBA9599,
    0xB8BDA50F,
    0x2802B89E,
    0x5F058808,
    0xC60CD9B2,
    0xB10BE924,
    0x2F6F7C87,
    0x58684C11,
    0xC1611DAB,
    0xB6662D3D,
    0x76DC4190,
    0x01DB7106,
    0x98D220BC,
    0xEFD5102A,
    0x71B18589,
    0x06B6B51F,
    0x9FBFE4A5,
    0xE8B8D433,
    0x7807C9A2,
    0x0F00F934,
    0x9609A88E,
    0xE10E9818,
    0x7F6A0DBB,
    0x086D3D2D,
    0x91646C97,
    0xE6635C01,
    0x6B6B51F4,
    0x1C6C6162,
    0x856530D8,
    0xF262004E,
    0x6C0695ED,
    0x1B01A57B,
    0x8208F4C1,
    0xF50FC457,
    0x65B0D9C6,
    0x12B7E950,
    0x8BBEB8EA,
    0xFCB9887C,
    0x62DD1DDF,
    0x15DA2D49,
    0x8CD37CF3,
    0xFBD44C65,
    0x4DB26158,
    0x3AB551CE,
    0xA3BC0074,
    0xD4BB30E2,
    0x4ADFA541,
    0x3DD895D7,
    0xA4D1C46D,
    0xD3D6F4FB,
    0x4369E96A,
    0x346ED9FC,
    0xAD678846,
    0xDA60B8D0,
    0x44042D73,
    0x33031DE5,
    0xAA0A4C5F,
    0xDD0D7CC9,
    0x5005713C,
    0x270241AA,
    0xBE0B1010,
    0xC90C2086,
    0x5768B525,
    0x206F85B3,
    0xB966D409,
    0xCE61E49F,
    0x5EDEF90E,
    0x29D9C998,
    0xB0D09822,
    0xC7D7A8B4,
    0x59B33D17,
    0x2EB40D81,
    0xB7BD5C3B,
    0xC0BA6CAD,
    0xEDB88320,
    0x9ABFB3B6,
    0x03B6E20C,
    0x74B1D29A,
    0xEAD54739,
    0x9DD277AF,
    0x04DB2615,
    0x73DC1683,
    0xE3630B12,
    0x94643B84,
    0x0D6D6A3E,
    0x7A6A5AA8,
    0xE40ECF0B,
    0x9309FF9D,
    0x0A00AE27,
    0x7D079EB1,
    0xF00F9344,
    0x8708A3D2,
    0x1E01F268,
    0x6906C2FE,
    0xF762575D,
    0x806567CB,
    0x196C3671,
    0x6E6B06E7,
    0xFED41B76,
    0x89D32BE0,
    0x10DA7A5A,
    0x67DD4ACC,
    0xF9B9DF6F,
    0x8EBEEFF9,
    0x17B7BE43,
    0x60B08ED5,
    0xD6D6A3E8,
    0xA1D1937E,
    0x38D8C2C4,
    0x4FDFF252,
    0xD1BB67F1,
    0xA6BC5767,
    0x3FB506DD,
    0x48B2364B,
    0xD80D2BDA,
    0xAF0A1B4C,
    0x36034AF6,
    0x41047A60,
    0xDF60EFC3,
    0xA867DF55,
    0x316E8EEF,
    0x4669BE79,
    0xCB61B38C,
    0xBC66831A,
    0x256FD2A0,
    0x5268E236,
    0xCC0C7795,
    0xBB0B4703,
    0x220216B9,
    0x5505262F,
    0xC5BA3BBE,
    0xB2BD0B28,
    0x2BB45A92,
    0x5CB36A04,
    0xC2D7FFA7,
    0xB5D0CF31,
    0x2CD99E8B,
    0x5BDEAE1D,
    0x9B64C2B0,
    0xEC63F226,
    0x756AA39C,
    0x026D930A,
    0x9C0906A9,
    0xEB0E363F,
    0x72076785,
    0x05005713,
    0x95BF4A82,
    0xE2B87A14,
    0x7BB12BAE,
    0x0CB61B38,
    0x92D28E9B,
    0xE5D5BE0D,
    0x7CDCEFB7,
    0x0BDBDF21,
    0x86D3D2D4,
    0xF1D4E242,
    0x68DDB3F8,
    0x1FDA836E,
    0x81BE16CD,
    0xF6B9265B,
    0x6FB077E1,
    0x18B74777,
    0x88085AE6,
    0xFF0F6A70,
    0x66063BCA,
    0x11010B5C,
    0x8F659EFF,
    0xF862AE69,
    0x616BFFD3,
    0x166CCF45,
    0xA00AE278,
    0xD70DD2EE,
    0x4E048354,
    0x3903B3C2,
    0xA7672661,
    0xD06016F7,
    0x4969474D,
    0x3E6E77DB,
    0xAED16A4A,
    0xD9D65ADC,
    0x40DF0B66,
    0x37D83BF0,
    0xA9BCAE53,
    0xDEBB9EC5,
    0x47B2CF7F,
    0x30B5FFE9,
    0xBDBDF21C,
    0xCABAC28A,
    0x53B39330,
    0x24B4A3A6,
    0xBAD03605,
    0xCDD70693,
    0x54DE5729,
    0x23D967BF,
    0xB3667A2E,
    0xC4614AB8,
    0x5D681B02,
    0x2A6F2B94,
    0xB40BBE37,
    0xC30C8EA1,
    0x5A05DF1B,
    0x2D02EF8D,
]


def att2idx(att: str) -> int | tuple[int]:
    """
    Get integer indices corresponding to grouped attribute.

    e.g. svid_06 -> 6; gnssId_103 -> 103, gsid_03_04 -> (3,4), tow -> 0

    :param str att: grouped attribute name e.g. svid_01
    :return: indices as integer(s), or 0 if not grouped
    :rtype: int | tuple[int]
    """

    try:
        att = att.split("_")
        ln = len(att)
        if ln == 2:  # one group level
            return int(att[1])
        if ln > 2:  # nested group level(s)
            return tuple(int(att[i]) for i in range(1, ln))
        return 0  # not grouped
    except ValueError:
        return 0


def att2name(att: str) -> str:
    """
    Get name of grouped attribute.

    e.g. svid_06 -> svid; gnssId_103 -> gnssId, tow -> tow

    :param str att: grouped attribute name e.g. svid_01
    :return: name without index e.g. svid
    :rtype: str
    """

    return att.split("_")[0]


def attsiz(att: str) -> int:
    """
    Helper function to return attribute size in bytes.

    :param str: attribute type e.g. 'U002'
    :return: size of attribute in bytes, or -1 if variable length
    :rtype: int

    """

    try:
        return int(att[1:4])
    except ValueError:
        return -1


def atttyp(att: str) -> str:
    """
    Helper function to return attribute type as string.

    :param str: attribute type e.g. 'U002'
    :return: type of attribute as string e.g. 'U'
    :rtype: str

    """

    return att[0:1]


def bytes2val(valb: bytes, att: str) -> Any:
    """
    Convert bytes to value for given UNI attribute type.

    :param bytes valb: attribute value in byte format e.g. b'\\\\x19\\\\x00\\\\x00\\\\x00'
    :param str att: attribute type e.g. 'U004'
    :return: attribute value as int, float, str or bytes
    :rtype: Any
    :raises: UNITypeError

    """

    if atttyp(att) == "X":  # bytes
        val = valb
    elif atttyp(att) == "C":  # string
        val = valb.decode("utf-8", errors="backslashreplace")
    elif atttyp(att) in ("S", "U"):  # integer
        val = int.from_bytes(valb, byteorder="little", signed=atttyp(att) == "S")
    elif atttyp(att) == "R":  # floating point
        val = struct.unpack("<f" if attsiz(att) == 4 else "<d", valb)[0]
    else:
        raise qge.UNITypeError(f"Unknown attribute type {att}")
    return val


def calc_crc(message: bytes) -> bytes:
    """
    Perform CRC32 cyclic redundancy check.

    TODO need to validate this algorithm

    :param bytes message: message
    :return: CRC as bytes
    :rtype: bytes

    """

    size = len(message)
    crc = 0
    for i in range(size):
        crc = CRCTABLE[(crc ^ message[i]) & 0xFF] ^ (crc >> 8)
    return val2bytes(crc, U4)

    # poly = 0x04C11DB7
    # bitmask = 0xFFFFFFFF
    # crc = 0
    # for byte in message:
    #     for _ in range(8):
    #         b = bitmask if byte & (1 << 7) != 0 else 0
    #         divide = bitmask if (crc & (1 << 31)) != 0 else 0
    #         crc = (crc << 1) ^ (poly & (b ^ divide))
    #         byte <<= 1
    #         crc &= bitmask
    # return crc


def escapeall(val: bytes) -> str:
    """
    Escape all byte characters e.g. b'\\\\x73' rather than b`s`

    :param bytes val: bytes
    :return: string of escaped bytes
    :rtype: str
    """

    return "b'{}'".format("".join(f"\\x{b:02x}" for b in val))


def get_bits(bitfield: bytes, bitmask: int) -> int:
    """
    Get integer value of specified (masked) bit(s) in a UNI bitfield (attribute type 'X')

    e.g. to get value of bits 6,7 in bitfield b'\\\\x89' (binary 0b10001001)::

        get_bits(b'\\x89', 0b11000000) = get_bits(b'\\x89', 192) = 2

    :param bytes bitfield: bitfield byte(s)
    :param int bitmask: bitmask as integer (= Σ(2**n), where n is the number of the bit)
    :return: value of masked bit(s)
    :rtype: int
    """

    i = 0
    val = int(bitfield.hex(), 16)
    while bitmask & 1 == 0:
        bitmask = bitmask >> 1
        i += 1
    return val >> i & bitmask


def isvalid_checksum(message: bytes) -> bool:
    """
    Validate message checksum.

    :param bytes message: message including header and checksum bytes
    :return: checksum valid flag
    :rtype: bool

    """

    lenm = len(message)
    ckm = message[lenm - 4 : lenm]
    return ckm == calc_crc(message[: lenm - 4])


def key_from_val(dictionary: dict, value) -> str:
    """
    Helper method - get dictionary key corresponding to (unique) value.

    :param dict dictionary: dictionary
    :param object value: unique dictionary value
    :return: dictionary key
    :rtype: str
    :raises: KeyError: if no key found for value

    """

    val = None
    for key, val in dictionary.items():
        if val == value:
            return key
    raise KeyError(f"No key found for value {value}")


def nomval(att: str) -> Any:
    """
    Get nominal value for given UNI attribute type.

    :param str att: attribute type e.g. 'U004'
    :return: attribute value as int, float, str or bytes
    :rtype: Any
    :raises: UNITypeError

    """

    if atttyp(att) == "X":
        val = b"\x00" * attsiz(att)
    elif atttyp(att) == "C":
        val = " " * attsiz(att)
    elif atttyp(att) == "R":
        val = 0.0
    elif atttyp(att) in ("S", "U"):
        val = 0
    else:
        raise qge.UNITypeError(f"Unknown attribute type {att}")
    return val


def val2bytes(val: Any, att: str) -> bytes:
    """
    Convert value to bytes for given UNI attribute type.

    :param Any val: attribute value e.g. 25
    :param str att: attribute type e.g. 'U004'
    :return: attribute value as bytes
    :rtype: bytes
    :raises: UNITypeError

    """

    try:
        if not isinstance(val, ATTTYPE[atttyp(att)]):
            raise TypeError(
                f"Attribute type {att} value {val} must be {ATTTYPE[atttyp(att)]}, not {type(val)}"
            )
    except KeyError as err:
        raise qge.UNITypeError(f"Unknown attribute type {att}") from err

    valb = val
    if atttyp(att) == "X":  # byte
        valb = val
    elif atttyp(att) == "C":  # string
        v = val.encode("utf-8", errors="backslashreplace")
        valb = v + b"\x20" * (attsiz(att) - len(v))  # right pad with spaces
    elif atttyp(att) in ("S", "U"):  # integer
        valb = val.to_bytes(attsiz(att), byteorder="little", signed=atttyp(att) == "S")
    elif atttyp(att) == "R":  # floating point
        valb = struct.pack("<f" if attsiz(att) == 4 else "<d", float(val))
    return valb


def utc2wnotow(utc: datetime = datetime.now(tz=timezone.utc)) -> tuple[int, int]:
    """
    Get GPS Week number (Wno) and Time of Week (Tow)
    in milliseconds for given utc datetime.

    GPS Epoch 0 = 6th Jan 1980

    :param datetime dat: calendar date
    :return: Wno, Tow
    :rtype: tuple[int,int]
    """

    ts = (utc - GPSEPOCH0).total_seconds() * 1000
    wno = int((utc - GPSEPOCH0).days / 7)
    tow = int(ts - wno * 604800000)
    return wno, tow


def timeinfo2vals(timeinfo: bytes) -> tuple:
    """
    Convert timeinfo bytes from header to values

    :param bytes timeinfo: timeinfo from header
    :return: individual values
    :rtype: tuple
    """

    timeref = bytes2val(timeinfo[0:1], U1)
    timestatus = bytes2val(timeinfo[1:2], U1)
    wno = bytes2val(timeinfo[2:4], U2)
    tow = bytes2val(timeinfo[4:8], U4)
    version = bytes2val(timeinfo[8:12], U4)
    # reserved = bytes2val(timeinfo[12:13], U1)
    leapsecond = bytes2val(timeinfo[13:14], U1)
    delay = bytes2val(timeinfo[14:16], U2)
    return timeref, timestatus, wno, tow, version, leapsecond, delay


def timeinfo2bytes(
    timeref: int = 1,
    timestatus: int = 0,
    wno: int | NoneType = None,
    tow: int | NoneType = None,
    version: int = 0,
    leapsecond: int = 0,
    delay: int = 0,
) -> bytes:
    """
    Convert timeinfo values to header bytes

    :param int timeref: Description
    :param int timestatus: Description
    :param int | NoneType wno: Description
    :param int | NoneType tow: Description
    :param int version: Description
    :param int leapsecond: Description
    :param int delay: Description
    :return: timeinfo as bytes
    :rtype: bytes
    """

    if wno is None or tow is None:
        wno, tow = utc2wnotow(datetime.now(tz=timezone.utc))

    return (
        val2bytes(timeref, U1)
        + val2bytes(timestatus, U1)
        + val2bytes(wno, U2)
        + val2bytes(tow, U4)
        + val2bytes(version, U4)
        + val2bytes(0, U1)
        + val2bytes(leapsecond, U1)
        + val2bytes(delay, U2)
    )



"""
UNI Protocol output definitions.

Created on 26 Jan 2026

Information sourced from public domain Unicore UM980 Interface Specifications © 2023, Unicore
https://www.ardusimple.com/wp-content/uploads/2023/04/Unicore-Reference-Commands-Manual-For-N4-High-Precision-Products_V2_EN_R1.4-1.pdf

:author: semuadmin (Steve Smith)
"""

from pyunigps.unitypes_core import (
    U2,
    U3,
    X1,
)

UNI_PAYLOADS_GET = {
    # TODO add payload definitions...
    # TODO order alphabetically
    "VERSION": {
        "device": "C004",
        "swversion": "C033",
        "authtype": "C129",
        "psn": "C066",
        "efuseid": "C033",
        "comptime": "C043",
    },
    "OBSVM": {},
    "OBSVH": {},
    "OBSVMCMP": {},
    "OBSVHCMP": {},
    "OBSVBASE": {},
    "BASEINFO": {},
    "GPSION": {},
    "BD3ION": {},
    "BDSION": {},
    "GALION": {},
    "GPSUTC": {},
    "BD3UTC": {},
    "BDSUTC": {},
    "GALUTC": {},
    "GPSEPH": {},
    "QZSSEPH": {},
    "BD3EPH": {},
    "BDSEPH": {},
    "GLOEPH": {},
    "GALEPH": {},
    "IRNSSEPH": {},
    "AGRIC": {},
    "PVTSLN": {},
    "UNILOGLIST": {},
    "BESTNAV": {},
    "BESTNAVXYZ": {},
    "BESTNAVH": {},
    "BESTNAVXYZH": {},
    "BESTSAT": {},
    "ADRNAV": {},
    "ADRNAVH": {},
    "PPPNAV": {},
    "SPPNAV": {},
    "SPPNAVH": {},
    "STADOP": {},
    "STADOPH": {},
    "ADRDOP": {},
    "ADRDOPH": {},
    "PPPDOP": {},
    "SPPDOP": {},
    "SPPDOPH": {},
    "SATSINFO": {},
    "BASEPOS": {},
    "SATELLITE": {},
    "SATECEF": {},
    "RECTIME": {},
    "UNIHEADING": {},
    "UNIHEADING2": {},
    "HEADINGSTATUS": {},
    "RTKSTATUS": {},
    "AGNSSSTATUS": {},
    "RTCSTATUS": {},
    "JAMSTATUS": {},
    "FREQJAMSTATUS": {},
    "RTCMSTATUS": {},
    "HWSTATUS": {},
    "AGC": {},
    "KSXT": {},
    "INFOPART1": {},
    "INFOPART2": {},
    "MSPOS": {},
    "TROPINFO": {},
    "PPPB2BINFO1": {},
    "PPPB2BINFO2": {},
    "PPPB2BINFO3": {},
    "PPPB2BINFO4": {},
    "PPPB2BINFO5": {},
    "PPPB2BINFO6": {},
    "PPPB2BINFO7": {},
    "E6MASKBLOCK": {},
    "E6ORBITBLOCK": {},
    "E6CLOCKFULLBLOCK": {},
    "E6CLOCKSUBBLOCK": {},
    "E6CBIASBLOCK": {},
    "E6PBIASBLOCK": {},
    "BSLNENUHD2": {},
    "BSLNXYZHD2": {},
    "DOPHD2": {},
    "TEST12": {"data": U3, "mode": U2},
    "TEST14": {"data": U3, "mode": U2, "status": U2},
    # ********************************************************************
    # UNI nominal payload definition, used as fallback where no documented
    # payload definition is available.
    # ********************************************************************
    "UNI-NOMINAL": {
        "group": (
            "None",
            {
                "data": X1,
            },
        )
    },
}



"""
UNIReader class.

Reads and parses individual UNI messages from any viable
data stream which supports a read(n) -> bytes method.

UNI message bit format (little-endian):

+----------+---------+---------+---------+-----------+----------+---------+
|   sync   | cpuidle |  msgid  | length  | timeinfo  | payload  |   crc   |
+==========+=========+=========+=========+===========+==========+=========+
| 0xaa44b5 | 1 byte  | 2 bytes | 2 bytes | 16 bytes  | variable | 4 bytes |
+----------+---------+---------+---------+-----------+----------+---------+
|                  header = 24 bytes                 |          |         |
+----------+---------+---------+---------+-----------+----------+---------+

timeinfo:

+---------+----------+---------+---------+---------+----------+---------+---------+
| timeref | timestat |   wno   |   tow   | version | reserved | leapsec |  delay  |
+=========+==========+=========+=========+=========+==========+=========+=========+
| 1 byte  |  1 byte  | 2 bytes | 4 bytes | 4 bytes |  1 byte  | 1 byte  | 2 bytes |
+---------+----------+---------+---------+---------+----------+---------+---------+

Returns both the raw binary data (as bytes) and the parsed data
(as an UNIMessage object).

- 'protfilter' governs which protocols (NMEA, UNI or RTCM3) are processed
- 'quitonerror' governs how errors are handled
- 'parsing' governs whether messages are fully parsed

Created on 26 Jan 2026

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2020
:license: BSD 3-Clause
"""

# pylint: disable=too-many-positional-arguments

from logging import getLogger
from socket import socket

import pynmeagps.exceptions as nme
import pyrtcm.exceptions as rte
from pynmeagps import NMEA_HDR, NMEAReader, SocketWrapper
from pyrtcm import RTCMReader

from pyunigps.exceptions import (
    UNIMessageError,
    UNIParseError,
    UNIStreamError,
    UNITypeError,
)
from pyunigps.unihelpers import (
    bytes2val,
    calc_crc,
    escapeall,
    val2bytes,
)
from pyunigps.unimessage import UNIMessage
from pyunigps.unitypes_core import (
    ERR_LOG,
    ERR_RAISE,
    GET,
    NMEA_PROTOCOL,
    POLL,
    RTCM3_PROTOCOL,
    SET,
    SETPOLL,
    U1,
    U2,
    U4,
    UNI_HDR,
    UNI_PROTOCOL,
    VALCKSUM,
)


class UNIReader:
    """
    UNIReader class.
    """

    def __init__(
        self,
        datastream,
        msgmode: int = GET,
        validate: int = VALCKSUM,
        protfilter: int = NMEA_PROTOCOL | UNI_PROTOCOL | RTCM3_PROTOCOL,
        quitonerror: int = ERR_LOG,
        parsebitfield: bool = True,
        bufsize: int = 4096,
        parsing: bool = True,
        errorhandler: object = None,
    ):
        """Constructor.

        :param datastream stream: input data stream
        :param int msgmode: 0=GET, 1=SET, 2=POLL, 3=SETPOLL (0)
        :param int validate: VALCKSUM (1) = Validate checksum,
            VALNONE (0) = ignore invalid checksum (1)
        :param int protfilter: NMEA_PROTOCOL (1), UNI_PROTOCOL (2), RTCM3_PROTOCOL (4),
            Can be OR'd (7)
        :param int quitonerror: ERR_IGNORE (0) = ignore errors,  ERR_LOG (1) = log continue,
            ERR_RAISE (2) = (re)raise (1)
        :param bool parsebitfield: 1 = parse bitfields, 0 = leave as bytes (1)
        :param int bufsize: socket recv buffer size (4096)
        :param bool parsing: True = parse data, False = don't parse data (output raw only) (True)
        :param object errorhandler: error handling object or function (None)
        :raises: UNIStreamError (if mode is invalid)
        """
        # pylint: disable=too-many-arguments

        if isinstance(datastream, socket):
            self._stream = SocketWrapper(datastream, bufsize=bufsize)
        else:
            self._stream = datastream
        self._protfilter = protfilter
        self._quitonerror = quitonerror
        self._errorhandler = errorhandler
        self._validate = validate
        self._parsebf = parsebitfield
        self._msgmode = msgmode
        self._parsing = parsing
        self._logger = getLogger(__name__)

        if self._msgmode not in (GET, SET, POLL, SETPOLL):
            raise UNIStreamError(
                f"Invalid stream mode {self._msgmode} - must be 0, 1, 2 or 3"
            )

    def __iter__(self):
        """Iterator."""

        return self

    def __next__(self) -> tuple:
        """
        Return next item in iteration.

        :return: tuple of (raw_data as bytes, parsed_data as UNIMessage)
        :rtype: tuple
        :raises: StopIteration

        """

        raw_data, parsed_data = self.read()
        if raw_data is None and parsed_data is None:
            raise StopIteration
        return (raw_data, parsed_data)

    def read(self) -> tuple:
        """
        Read a single UNI message from the stream buffer
        and return both raw and parsed data.

        'quitonerror' determines whether to raise, log or ignore parsing errors.

        :return: tuple of (raw_data as bytes, parsed_data as UNIMessage)
        :rtype: tuple
        :raises: Exception (if invalid or unrecognised protocol in data stream)
        """

        parsing = True
        while parsing:  # loop until end of valid message or EOF
            try:

                raw_data = None
                parsed_data = None
                byte1 = self._read_bytes(1)  # read the first byte
                # if not UNI, NMEA or RTCM3, discard and continue
                if byte1 not in (b"\xaa", b"\x24", b"\xd3"):
                    continue
                byte2 = self._read_bytes(1)
                bytehdr = byte1 + byte2
                if bytehdr == UNI_HDR[0:2]:
                    byte3 = self._read_bytes(1)
                    bytehdr += byte3
                    # if it's a UNI message (b'\xaa\x44\b5')
                    if bytehdr != UNI_HDR:
                        continue
                    raw_data, parsed_data = self._parse_uni(bytehdr)
                    # if protocol filter passes UNI, return message,
                    # otherwise discard and continue
                    if self._protfilter & UNI_PROTOCOL:
                        parsing = False
                    else:
                        continue
                # if it's an NMEA message (b'\x24\x..)
                elif bytehdr in NMEA_HDR:
                    raw_data, parsed_data = self._parse_nmea(bytehdr)
                    # if protocol filter passes NMEA, return message,
                    # otherwise discard and continue
                    if self._protfilter & NMEA_PROTOCOL:
                        parsing = False
                    else:
                        continue
                # if it's a RTCM3 message
                # (byte1 = 0xd3; byte2 = 0b000000**)
                elif byte1 == b"\xd3" and (byte2[0] & ~0x03) == 0:
                    raw_data, parsed_data = self._parse_rtcm3(bytehdr)
                    # if protocol filter passes RTCM, return message,
                    # otherwise discard and continue
                    if self._protfilter & RTCM3_PROTOCOL:
                        parsing = False
                    else:
                        continue
                # unrecognised protocol header
                else:
                    raise UNIParseError(f"Unknown protocol header {bytehdr}.")

            except EOFError:
                return (None, None)
            except (
                UNIMessageError,
                UNITypeError,
                UNIParseError,
                UNIStreamError,
                nme.NMEAMessageError,
                nme.NMEATypeError,
                nme.NMEAParseError,
                nme.NMEAStreamError,
                rte.RTCMMessageError,
                rte.RTCMParseError,
                rte.RTCMStreamError,
                rte.RTCMTypeError,
            ) as err:
                if self._quitonerror:
                    self._do_error(err)
                continue

        return (raw_data, parsed_data)

    def _parse_uni(self, hdr: bytes) -> tuple:
        """
        Parse remainder of UNI message.

        :param bytes hdr: UNI header (b'\\xaa\\x44\\xb5')
        :return: tuple of (raw_data as bytes, parsed_data as UNIMessage or None)
        :rtype: tuple
        """

        # read the rest of the UNI message from the buffer
        byten = self._read_bytes(21)
        cpuidle = byten[0:1]
        msgid = byten[1:3]
        lenb = byten[3:5]
        timeinfo = byten[5:21]
        leni = int.from_bytes(lenb, "little", signed=False)
        byten = self._read_bytes(leni + 4)
        plb = byten[0:leni]
        crc = byten[leni : leni + 4]
        raw_data = hdr + cpuidle + msgid + lenb + timeinfo + plb + crc
        # only parse if we need to (filter passes UNI)
        if (self._protfilter & UNI_PROTOCOL) and self._parsing:
            parsed_data = self.parse(
                raw_data,
                msgmode=self._msgmode,
                validate=self._validate,
                parsebitfield=self._parsebf,
            )
        else:
            parsed_data = None
        return (raw_data, parsed_data)

    def _parse_nmea(self, hdr: bytes) -> tuple:
        """
        Parse remainder of NMEA message (using pynmeagps library).

        :param bytes hdr: NMEA header (b'\\x24\\x..')
        :return: tuple of (raw_data as bytes, parsed_data as NMEAMessage or None)
        :rtype: tuple
        """

        # read the rest of the NMEA message from the buffer
        byten = self._read_line()  # NMEA protocol is CRLF-terminated
        raw_data = hdr + byten
        # only parse if we need to (filter passes NMEA)
        if (self._protfilter & NMEA_PROTOCOL) and self._parsing:
            # invoke pynmeagps parser
            parsed_data = NMEAReader.parse(
                raw_data,
                validate=self._validate,
                msgmode=self._msgmode,
            )
        else:
            parsed_data = None
        return (raw_data, parsed_data)

    def _parse_rtcm3(self, hdr: bytes) -> tuple:
        """
        Parse any RTCM3 data in the stream (using pyrtcm library).

        :param bytes hdr: first 2 bytes of RTCM3 header
        :return: tuple of (raw_data as bytes, parsed_stub as RTCMMessage)
        :rtype: tuple
        """

        hdr3 = self._read_bytes(1)
        size = hdr3[0] | (hdr[1] << 8)
        payload = self._read_bytes(size)
        crc = self._read_bytes(3)
        raw_data = hdr + hdr3 + payload + crc
        # only parse if we need to (filter passes RTCM)
        if (self._protfilter & RTCM3_PROTOCOL) and self._parsing:
            # invoke pyrtcm parser
            parsed_data = RTCMReader.parse(
                raw_data,
                validate=self._validate,
                labelmsm=1,
            )
        else:
            parsed_data = None
        return (raw_data, parsed_data)

    def _read_bytes(self, size: int) -> bytes:
        """
        Read a specified number of bytes from stream.

        :param int size: number of bytes to read
        :return: bytes
        :rtype: bytes
        :raises: UNIStreamError if stream ends prematurely
        """

        data = self._stream.read(size)
        if len(data) == 0:  # EOF
            raise EOFError()
        if 0 < len(data) < size:  # truncated stream
            raise UNIStreamError(
                "Serial stream terminated unexpectedly. "
                f"{size} bytes requested, {len(data)} bytes returned."
            )
        return data

    def _read_line(self) -> bytes:
        """
        Read bytes until LF (0x0a) terminator.

        :return: bytes
        :rtype: bytes
        :raises: UNIStreamError if stream ends prematurely
        """

        data = self._stream.readline()  # NMEA protocol is CRLF-terminated
        if len(data) == 0:
            raise EOFError()  # pragma: no cover
        if data[-1:] != b"\x0a":  # truncated stream
            raise UNIStreamError(
                "Serial stream terminated unexpectedly. "
                f"Line requested, {len(data)} bytes returned."
            )
        return data

    def _do_error(self, err: Exception):
        """
        Handle error.

        :param Exception err: error
        :raises: Exception if quitonerror = ERR_RAISE (2)
        """

        if self._quitonerror == ERR_RAISE:
            raise err from err
        if self._quitonerror == ERR_LOG:
            # pass to error handler if there is one
            # else just log
            if self._errorhandler is None:
                self._logger.error(err)
            else:
                self._errorhandler(err)

    @property
    def datastream(self) -> object:
        """
        Getter for stream.

        :return: data stream
        :rtype: object
        """

        return self._stream

    @staticmethod
    def parse(
        message: bytes,
        msgmode: int = GET,
        validate: int = VALCKSUM,
        parsebitfield: bool = True,
    ) -> object:
        """
        Parse UNI byte stream to UNIMessage object.

        :param bytes message: binary message to parse
        :param int msgmode: GET (0), SET (1), POLL (2) (0)
        :param int validate: VALCKSUM (1) = Validate checksum,
            VALNONE (0) = ignore invalid checksum (1)
        :param bool parsebitfield: 1 = parse bitfields, 0 = leave as bytes (1)
        :return: UNIMessage object
        :rtype: UNIMessage
        :raises: Exception (if data stream contains invalid data or unknown message type)
        """

        if msgmode not in (GET, SET, POLL, SETPOLL):
            raise UNIParseError(
                f"Invalid message mode {msgmode} - must be 0, 1, 2 or 3"
            )

        lenm = len(message)
        hdr = message[0:3]
        cpuidleb = message[3:4]
        cpuidle = bytes2val(cpuidleb, U1)
        msgidb = message[4:6]
        msgid = bytes2val(msgidb, U2)
        lenb = message[6:8]
        length = bytes2val(lenb, U2)
        timeref = bytes2val(message[8:9], U1)
        timestatus = bytes2val(message[9:10], U1)
        wno = bytes2val(message[10:12], U2)
        tow = bytes2val(message[12:16], U4)
        version = bytes2val(message[16:20], U4)
        leapsecond = bytes2val(message[21:22], U1)
        delay = bytes2val(message[22:24], U1)
        crcb = message[lenm - 4 : lenm]

        if lenb == b"\x00\x00\x00\x00":
            payload = None
            lenp = 0
        else:
            payload = message[24 : lenm - 4]
            lenp = len(payload)

        if payload is None:
            crc = calc_crc(message[0:24])
        else:
            crc = calc_crc(message[0:24] + payload)

        if validate & VALCKSUM:
            if hdr != UNI_HDR:
                raise UNIParseError(
                    (
                        f"Invalid message header {escapeall(hdr)}"
                        f" - should be {escapeall(UNI_HDR)}"
                    )
                )
            if lenp != length:
                raise UNIParseError(
                    (
                        f"Invalid payload length {escapeall(lenb)}"
                        f" - should be {val2bytes(lenp, U2)}"
                    )
                )
            if crc != crcb:
                raise UNIParseError(
                    (
                        f"Message checksum {escapeall(crcb)}"
                        f" invalid - should be {escapeall(crc)}"
                    )
                )
        parsed_data = UNIMessage(
            msgid=msgid,
            length=length,
            cpuidle=cpuidle,
            timeref=timeref,
            timestatus=timestatus,
            wno=wno,
            tow=tow,
            version=version,
            leapsecond=leapsecond,
            delay=delay,
            checksum=crcb,
            msgmode=msgmode,
            parsebitfield=parsebitfield,
            payload=payload,
        )
        return parsed_data



"""
UNI Custom Exception Types.

Created on 6 Oct 2025

:author: semuadmin (Steve Smith)
:copyright: semuadmin © 2020
:license: BSD 3-Clause
"""


class ParameterError(Exception):
    """Parameter Error Class."""


class GNSSStreamError(Exception):
    """Generic Stream Error Class."""


class UNIParseError(Exception):
    """
    UNI Parsing error.
    """


class UNIStreamError(Exception):
    """
    UNI Streaming error.
    """


class UNIMessageError(Exception):
    """
    UNI Undefined message class/id.
    Essentially a prompt to add missing payload types to UNI_PAYLOADS.
    """


class UNITypeError(Exception):
    """
    UNI Undefined payload attribute type.
    Essentially a prompt to fix incorrect payload definitions to UNI_PAYLOADS.
    """
